#ifndef _BOOST_QLANG_DETAIL_DETAIL_HPP_
#define _BOOST_QLANG_DETAIL_DETAIL_HPP_


#include <boost/mpl/if.hpp>
#include <boost/type_traits/is_same.hpp>
#include <boost/iterator/filter_iterator.hpp>
#include <boost/iterator/transform_iterator.hpp>
#include <boost/bind.hpp>
#include <boost/tuple/tuple.hpp>
#include <boost/function.hpp>
#include <boost/range/iterator_range.hpp>


namespace boost  {
namespace qlang {


template <class Range, class Predicate, class Placeholder>
struct select_;

struct full_object_t {};

namespace detail {

/**********************DECLARATIONS**********************/
template <class T>
struct extract_ret_type_member_func;


template<class T> 
struct remove_mem_ptr;


template <class T>
struct TruePredicate;

template <class T>
struct identity;


template <bool cond>
struct restart_traversal;

template <class T>
struct is_query;

template <class ObjectType, class MemPtr>
struct field_select;


template <class Range, class Predicate, class MemPtr> 
struct transformed_filtered_range_type;


template <class Range, class Placeholder>
struct unary_function_type;


template <class Range,
          class TransformedFilteredRange, class Predicate, class Placeholder>
struct select_init_range;




/***********************************IMPLEMENTATION****************************/
template <class T>
struct extract_ret_type_member_func {};


template <class Obj, class R, class... Args>
struct extract_ret_type_member_func<R (Obj::*)(Args...)>
{
    typedef R type;
};


template <class Obj, class R, class... Args>
struct extract_ret_type_member_func<R (Obj::*)(Args...) const>
{
    typedef R type;
};



template<class T> struct remove_mem_ptr {
    typedef T type;
};

template<class T, class In> struct remove_mem_ptr<T In::*> {
    typedef T type;
}; 



template <class T>
struct TruePredicate {
    bool operator()(const T & t) const { return true; }
};



template <class T>
struct identity : std::unary_function<const T &, T>
{
    T operator()(const T & t) const { return t; }
};


template <class ObjectType, class MemPtr>
struct field_select : std::binary_function<MemPtr, const ObjectType &,
const boost::tuple<typename detail::remove_mem_ptr<MemPtr>::type> >
{
    const boost::tuple<typename detail::remove_mem_ptr<MemPtr>::type>
    operator()(MemPtr mem_ptr, const ObjectType & val) const {
        return boost::make_tuple(boost::bind(mem_ptr, ::_1)(val));      
    } 
};



template <class Range, class Predicate, class MemPtr> 
struct transformed_filtered_range_type 
: 
boost::mpl::if_
    <   boost::is_same<full_object_t, MemPtr>
        ,
        boost::iterator_range<
        boost::filter_iterator<Predicate, 
        typename Range::const_iterator> >
        , 
        boost::iterator_range<
        boost::transform_iterator<
        boost::function<const boost::tuple<typename detail::remove_mem_ptr<MemPtr>::type> 
        (const typename Range::value_type &)>,
        boost::filter_iterator<Predicate, typename Range::const_iterator> > >
    >
{};



template <class Range, class Placeholder>
struct unary_function_type 
:
boost::mpl::if_
    <   boost::is_same<full_object_t, Placeholder>
        ,
        boost::function<typename Range::value_type (const typename Range::value_type &)>
        ,
        boost::function<const boost::tuple<typename detail::remove_mem_ptr<Placeholder>::type> 
        (const typename Range::value_type &)> >
{};



template <class Range,
class TransformedFilteredRange, class Predicate, class Placeholder>
struct select_init_range
{
    
    TransformedFilteredRange operator()(const Range & range, const Predicate & pred, 
    Placeholder ph) const
    {
        
        return TransformedFilteredRange(
        boost::make_transform_iterator(boost::make_filter_iterator(pred, boost::begin(range), boost::end(range)),
        boost::bind(detail::field_select<typename Range::value_type, Placeholder>(), ph, ::_1)),
        boost::make_transform_iterator(boost::make_filter_iterator(pred, boost::end(range), boost::end(range)),
        boost::bind(detail::field_select<typename Range::value_type, Placeholder>(), ph, ::_1)));
        
    }
};

template <class Range,
class TransformedFilteredRange, class Predicate>
struct select_init_range<Range, TransformedFilteredRange, Predicate, full_object_t>
{
    
    TransformedFilteredRange operator()(const Range & range, const Predicate & pred, 
    full_object_t ph) const
    {
        return TransformedFilteredRange(
        boost::make_filter_iterator(pred, boost::begin(range), boost::end(range)),
        boost::make_filter_iterator(pred, boost::end(range), boost::end(range)));
    }
};


/////////////////////////////////////////////////////////////////////////////////////////////////////
template <class T>
struct is_query : boost::false_type {};

template <class R, class Pred, class Pholder>
struct is_query<select_<R, Pred, Pholder> > : boost::true_type {};


template <bool cond>
struct restart_traversal {
    template <class Range>
    void operator()(const Range & r) const { const_cast<Range&>(r).restart_traversal(); }
};


template <>
struct restart_traversal<false> {
    template <class Range>
    void operator()(const Range & r) const {}
};


} //namespace detail

} //namespace qlang

} //namespace boost


#endif
