#ifndef _BOOST_QLANG_SELECT_HPP_
#define _BOOST_QLANG_SELECT_HPP_

#include <boost/qlang/detail/detail.hpp>
#include <boost/range/begin.hpp>
#include <boost/range/end.hpp>
#include <vector>
#include <list>


namespace boost {
namespace qlang {


template <class Range, class Predicate, class Placeholder>
class select_ {
    typedef typename detail::transformed_filtered_range_type<Range, Predicate, Placeholder>::type transformed_filtered_range;
    typedef typename detail::unary_function_type<Range, Placeholder>::type UnaryFunc;



    const Range & range_;
    const Predicate pred_;
    Placeholder ph_;
    transformed_filtered_range frange_;

    friend class detail::restart_traversal<false>;
    friend class detail::restart_traversal<true>;

    void restart_traversal();
public:
    typedef typename Range::value_type value_type;
    typedef typename transformed_filtered_range::const_iterator const_iterator;
    typedef typename transformed_filtered_range::const_iterator       iterator;
    
    typename transformed_filtered_range::const_iterator begin() const; 
    typename transformed_filtered_range::const_iterator end() const;
    

    select_(const Range & rng, const Predicate & pred, Placeholder ph) : range_(rng),
        pred_(pred), ph_(ph),
        frange_(
        detail::
        select_init_range<Range, transformed_filtered_range, Predicate, Placeholder>()
        (range_, pred_, ph_)) {}

    
    std::vector<typename transformed_filtered_range::value_type> to_vector() const {
        std::vector<typename transformed_filtered_range::value_type> vec;
        std::copy(boost::begin(*this), boost::end(*this), back_inserter(vec));
        return vec;
    }

    std::list<typename transformed_filtered_range::value_type> to_list() const {
        std::list<typename transformed_filtered_range::value_type> l;
        std::copy(boost::begin(*this), boost::end(*this), back_inserter(l));
        return l;
    }
    
};




template <class Range, class Predicate, class Placeholder>
typename select_<Range, Predicate, Placeholder>::transformed_filtered_range::const_iterator 
select_<Range, Predicate, Placeholder>::begin() const
{ 
    const_cast<select_<Range, Predicate, Placeholder>*>(this)->restart_traversal();
    return boost::begin(this->frange_);    
}



template <class Range, class Predicate, class Placeholder>
typename select_<Range, Predicate, Placeholder>::transformed_filtered_range::const_iterator 
select_<Range, Predicate, Placeholder>::end() const
{ 
    return boost::end(this->frange_);
}


template <class Range, class Predicate, class Placeholder>
void select_<Range, Predicate, Placeholder>::restart_traversal() {
    detail::restart_traversal<detail::is_query<Range>::value>()(range_);

    frange_ = detail::
    select_init_range<Range, transformed_filtered_range, Predicate, Placeholder>()
    (range_, pred_, ph_);
    
}



} //namespace qlang

} //namespace boost


#endif
