#include <boost/qlang/qlang.hpp>
#include <vector>
#include <iostream>
#include <iterator>



struct Person {
    std::string name_;
    std::string surname_;
    int age_;

public:
    Person(const std::string & name, const std::string & surname,
           int age) :
    name_(name), surname_(surname), age_(age) {}
};



using namespace boost::qlang;
using namespace std;



int main(int argc, char * argv[])
{
    vector<Person> persons = { {"German", "Gallardone", 23}, {"Pedro", "Ruiz", 28},
                                {"Hilario", "Pinose", 34}, {"Manolito", "Gafotas", 29},
                                {"Tomas",  "Todonte", 56}, {"Tomas", "Ruiz", 33}};


    //Constructs a tuple with Person::surname_ of type std::string (like surname)
    auto query = from(persons).where([](const Person & d) { return d.surname_ == "Ruiz"; }).
                     select(&Person::surname_);

    //Selects the original object
    auto query2  = from(persons).where([](const Person & d) { return d.surname_ == "Ruiz"; }).
                     select(full_object);


    //Nested query: Uses query2 and filters by age
    auto query3 = from(query2).where([](const Person & a) { return a.age_ == 28; }).
                     select(&Person::name_);

    //Simple query that returns every object from persons
    auto query4  = from(persons).select(full_object);


    for (auto it = query.begin(); it != query.end(); ++it)
    {
        cout << get<0>(*it) << endl;
    }    
    cout << endl;


    //When push_back is used, if you execute a query again, it will consider this new
    //inserted element, since queries are lazy.
    persons.push_back({"Manolo", "Ruiz", 43});

    //When using begin and end, the query is evaluated as it runs
    for (auto it = query2.begin(); it != query2.end(); ++it)
    {
        cout << it->name_ << endl;
    }
    cout << endl;


    //When using begin and end, the query is evaluated as it runs
    for (auto it = query3.begin(); it != query3.end(); ++it)
    {
        cout << get<0>(*it) << endl;
    }

    cout << endl;

    //When using begin and end, the query is evaluated as it runs
    for (auto it = query4.begin(); it != query4.end(); ++it)
    {
        cout << it->name_ << endl;
    }
}


